#  coding=utf-8
#  Copyright 2023 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import re
from distutils.core import setup
from platform import system, machine, python_version
from setuptools import find_namespace_packages
from sys import version_info as pyversion

# Ensure we match the version set in optimum/nvidia/version.py
filepath = "src/optimum/nvidia/version.py"
try:
    with open(filepath) as version_file:
        (__version__,) = re.findall('__version__ = "(.*)"', version_file.read())
except Exception as error:
    assert False, "Error: Could not open '%s' due %s\n" % (filepath, error)

INSTALL_REQUIRES = [
    "accelerate == 0.25",
    "datasets >= 2.14",
    "huggingface-hub >= 0.22.0",
    "hf-transfer",
    "mpmath == 1.3.0",
    "numpy >= 1.26.0",
    "onnx >= 1.12.0",
    "optimum >= 1.13.0",
    "setuptools",
    "tensorrt-llm == 0.9.0.dev2024031900",
    "torch>=2.2.0a,<=2.3.0a",
    "transformers >= 4.38.2",
    "pynvml"
]

TESTS_REQUIRES = [
    "mock",
    "pytest",
    "pytest-xdist",
    "psutil",
    "parameterized",
    "datasets",
    "safetensors",
    "soundfile",
    "librosa",
]

QUALITY_REQUIRES = [
    "black",
    "ruff",
    "isort",
    "hf_doc_builder @ git+https://github.com/huggingface/doc-builder.git",
]


EXTRAS_REQUIRE = {
    "tests": TESTS_REQUIRES,
#     "quality": QUALITY_REQUIRES,
}

setup(
    name="optimum-nvidia",
    version=__version__,
    description=(
        "Optimum Nvidia is the interface between the Hugging Face Transformers and NVIDIA GPUs. "
        "It provides a set of tools enabling easy model loading, training and "
        "inference on single and multiple GPU cards for different downstream tasks."
    ),
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="transformers, neural-network, fine-tuning, inference, nvidia, tensorrt, ampere, hopper",
    url="https://huggingface.co/hardware/nvidia",
    author="HuggingFace Inc. Machine Learning Optimization Team",
    author_email="hardware@huggingface.co",
    license="Apache 2.0",
    packages=find_namespace_packages(include=["optimum*"]),
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    dependency_links=["https://pypi.nvidia.com"],
    include_package_data=True,
    zip_safe=False,
)



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from argparse import ArgumentParser, Namespace

from tensorrt_llm.quantization import QuantMode


# Model topology (sharding, pipelining, dtype)
def register_common_model_topology_args(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument(
        "--tensor-parallelism",
        type=int,
        default=1,
        help="Define the number of slice for each tensor each GPU will receive.",
    )
    parser.add_argument(
        "--pipeline-parallelism",
        type=int,
        default=1,
        help="Define the number of sections to split neural network layers",
    )
    parser.add_argument(
        "--world-size",
        type=int,
        default=1,
        help="Total number of GPUs over all the nodes.",
    )
    parser.add_argument(
        "--gpus-per-node",
        type=int,
        default=1,
        help="Total number of GPUs on a single node.",
    )
    parser.add_argument(
        "-o",
        "--opt-level",
        type=int,
        default=2,
        dest="optimization_level",
        help="Optimization level between 0 (no optimization) and 5 (maximum level of optimization).",
    )
    return parser


# TensorRT's optimization profiles
def register_optimization_profiles_args(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument(
        "--max-batch-size",
        type=int,
        default=1,
        help="Maximum batch size for the model.",
    )
    parser.add_argument(
        "--max-prompt-length",
        type=int,
        default=128,
        help="Maximum prompt a user can give.",
    )
    parser.add_argument(
        "--max-new-tokens",
        type=int,
        default=1024,
        help="Maximum number of tokens to generate",
    )
    parser.add_argument(
        "--max-beam-width",
        type=int,
        default=1,
        help="Maximum number of beams for sampling",
    )

    return parser


# Triton Inference Server
def register_triton_server_args(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument(
        "--with-triton-structure",
        action="store_true",
        help="Generate the Triton Inference Server structure",
    )
    return parser


# Quantization
def register_quantization_args(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument(
        "--fp8", action="store_true", help="Enable FP8 quantization for Ada & Hopper."
    )
    parser.add_argument(
        "--fp8-cache",
        action="store_true",
        help="Enable KV cache as fp8 for Ada & Hopper.",
    )
    parser.add_argument(
        "--num-calibration-samples",
        type=int,
        default=512,
        help="How much samples to use when calibrating.",
    )
    return parser


def postprocess_quantization_parameters(params: Namespace) -> Namespace:
    # Only support FP8 quantization for now
    qconfig = QuantMode.from_description(
        quantize_weights=False,
        quantize_activations=False,
        per_token=False,
        per_channel=False,
        per_group=False,
        use_int4_weights=False,
        use_int8_kv_cache=False,
        use_fp8_kv_cache=params.fp8_cache,
        use_fp8_qdq=params.fp8,
    )

    params.has_quantization_step = qconfig != QuantMode(0)
    params.quantization_config = qconfig

    # If we do have the output path, then let's create the calibration path
    if "output" in params:
        from pathlib import Path

        params.calibration_output = Path(params.output).joinpath("calibration")

    return params



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from os import PathLike
from typing import Dict, Optional, Tuple, Type, Union

from huggingface_hub import model_info
from tensorrt_llm import Module
from transformers import AutoTokenizer, PreTrainedTokenizer, PreTrainedTokenizerFast

from optimum.nvidia import AutoModelForCausalLM
from optimum.nvidia.pipelines.text_generation import TextGenerationPipeline

from .base import Pipeline


SUPPORTED_MODEL_WITH_TASKS: Dict[str, Dict[str, Tuple[Type[Pipeline], Type]]] = {
    "gemma": {"text-generation": (TextGenerationPipeline, AutoModelForCausalLM)},
    "llama": {"text-generation": (TextGenerationPipeline, AutoModelForCausalLM)},
    "mistral": {"text-generation": (TextGenerationPipeline, AutoModelForCausalLM)},
}


def get_target_class_for_model_and_task(task: str, architecture: str) -> Optional[Type]:
    task_ = SUPPORTED_MODEL_WITH_TASKS.get(task, None)
    if not task:
        raise NotImplementedError(f"Task {task} is not supported yet.")

    target = task_.get(architecture, None)

    if not target:
        raise NotImplementedError(
            f"Architecture {architecture} is not supported for task {task}. "
            f"Only the following architectures are: {list(task_.keys())}"
        )

    return target


def pipeline(
    task: str = None,
    model: Union[str, PathLike, Module] = None,
    tokenizer: Optional[
        Union[str, PreTrainedTokenizer, PreTrainedTokenizerFast]
    ] = None,
    **kwargs,
):
    """
    Utility factory method to build a [`Pipeline`].

    Pipelines are made of:

        - A [tokenizer](tokenizer) in charge of mapping raw textual input to token.
        - A [model](model) to make predictions from the inputs.
        - Some (optional) post processing for enhancing model's output.

    Args:
        task (`str`):
            The task defining which pipeline will be returned. Currently accepted tasks are:
                - `"text-generation"`: will return a [`TextGenerationPipeline`]:.
        model (`str` or [`PreTrainedModel`] or [`TFPreTrainedModel`], *optional*):
            The model that will be used by the pipeline to make predictions. This can be a model identifier or an
            actual instance of a pretrained model inheriting from [`PreTrainedModel`] (for PyTorch) or
            [`TFPreTrainedModel`] (for TensorFlow).

            If not provided, the default for the `task` will be loaded.
        tokenizer (`str` or [`PreTrainedTokenizer`], *optional*):
            The tokenizer that will be used by the pipeline to encode data for the model. This can be a model
            identifier or an actual pretrained tokenizer inheriting from [`PreTrainedTokenizer`].

            If not provided, the default tokenizer for the given `model` will be loaded (if it is a string). If `model`
            is not specified or not a string, then the default tokenizer for `config` is loaded (if it is a string).
            However, if `config` is also not given or not a string, then the default tokenizer for the given `task`
            will be loaded.

    """

    try:
        info = model_info(model)
    except Exception as e:
        raise RuntimeError(
            f"Failed to instantiate the pipeline inferring the task for model {model}: {e}"
        )

    # Retrieve the model type
    model_type = info.config.get("model_type", None)
    if not model_type:
        raise RuntimeError(f"Failed to infer model type for model {model}")
    elif model_type not in SUPPORTED_MODEL_WITH_TASKS:
        raise NotImplementedError(f"Model type {model_type} is not currently supported")

    if not task and getattr(info, "library_name", "transformers") == "transformers":
        if not info.pipeline_tag:
            raise RuntimeError(
                f"Failed to infer the task for model {model}, please use `task` parameter"
            )
        task = info.pipeline_tag

    if task not in SUPPORTED_MODEL_WITH_TASKS[model_type]:
        raise NotImplementedError(f"Task {task} is not supported yet for {model_type}.")

    if tokenizer is None:
        tokenizer = AutoTokenizer.from_pretrained(model, use_fast=True)

    # Allocate
    pipeline_factory, model_factory = SUPPORTED_MODEL_WITH_TASKS[model_type][task]
    model = model_factory.from_pretrained(model, **kwargs)

    return pipeline_factory(model, tokenizer)



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path
from typing import Any, Dict, Optional, Type, Union

from huggingface_hub import ModelHubMixin

from optimum.nvidia.errors import UnsupportedModelException

from ..hub import extract_model_type
from .gemma import GemmaForCausalLM
from .llama import LlamaForCausalLM
from .mistral import MistralForCausalLM


_SUPPORTED_MODEL_CLASS = {
    "llama": LlamaForCausalLM,
    "mistral": MistralForCausalLM,
    "gemma": GemmaForCausalLM,
}


class AutoModelForCausalLM(ModelHubMixin):
    """ """

    @classmethod
    def _from_pretrained(
        cls: Type,
        *,
        model_id: str,
        revision: Optional[str],
        cache_dir: Optional[Union[str, Path]],
        force_download: bool,
        proxies: Optional[Dict],
        resume_download: bool,
        local_files_only: bool,
        token: Optional[Union[str, bool]],
        config: Optional[Dict[str, Any]] = None,
        **model_kwargs,
    ):
        if config is None:
            raise ValueError("Unable to determine the model type with config = None")

        model_type, _ = extract_model_type(config)
        if model_type is None:
            raise ValueError(
                "Unable to determine the model type from the provided config. "
                "Please open-up an issue at https://github.com/huggingface/optimum-nvidia/issues"
            )

        if model_type not in _SUPPORTED_MODEL_CLASS:
            raise UnsupportedModelException(model_type)

        model_clazz = _SUPPORTED_MODEL_CLASS[model_type]
        model = model_clazz.from_pretrained(
            pretrained_model_name_or_path=model_id,
            config=config,
            revision=revision,
            cache_dir=cache_dir,
            force_download=force_download,
            proxies=proxies,
            resume_download=resume_download,
            local_files_only=local_files_only,
            token=token,
            **model_kwargs,
        )

        return model



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from logging import getLogger
from typing import Dict

import numpy as np
from tensorrt_llm.models import PretrainedConfig, PretrainedModel
from tensorrt_llm.models.llama.model import LLaMAForCausalLM
from tensorrt_llm.models.llama.weight import load_from_hf_llama
from tensorrt_llm.plugin import PluginConfig
from transformers import MistralForCausalLM as TransformersMistralForCausalLM
from transformers import PretrainedConfig as TransformersPretrainedConfig
from transformers import PreTrainedModel as TransformersPretrainedModel

from optimum.nvidia import TensorRTConfig
from optimum.nvidia.config import dtype_to_str
from optimum.nvidia.hub import HuggingFaceHubModel
from optimum.nvidia.runtime import CausalLM


LOGGER = getLogger(__name__)


class MistralConfig(TensorRTConfig):
    r"""
    This is the configuration class to store the configuration of a [`LlamaModel`]. It is used to instantiate an LLaMA
    model according to the specified arguments, defining the model architecture. Instantiating a configuration with the
    defaults will yield a similar configuration to that of the LLaMA-7B.

    Configuration objects inherit from [`TensorRTConfig`] and can be used to control the model outputs. Read the
    documentation from [`TensorRTConfig`] for more information.
    """

    @staticmethod
    def from_config(config: TransformersPretrainedConfig) -> "TensorRTConfig":
        # Retrieve the quantization from the transformers config (if provided)
        _, qconfig = TensorRTConfig.get_quantization_config(config)

        trt_config = MistralConfig(
            architecture=config.architectures[0],
            dtype=dtype_to_str(config.torch_dtype),
            logits_dtype="float32",
            vocab_size=config.vocab_size,
            max_position_embeddings=config.max_position_embeddings,
            hidden_size=config.hidden_size,
            num_hidden_layers=config.num_hidden_layers,
            num_attention_heads=config.num_attention_heads,
            num_key_value_heads=getattr(
                config, "num_key_value_heads", config.num_attention_heads
            ),
            hidden_act=config.hidden_act,
            intermediate_size=config.intermediate_size,
            norm_epsilon=config.rms_norm_eps,
            position_embedding_type="rope_gpt_neox",
            world_size=1,
            tp_size=1,
            pp_size=1,
            use_prompt_tuning=False,
            use_parallel_embedding=False,
            embedding_sharding_dim=0,
            share_embedding_table=False,
            max_lora_rank=64,
            head_size=config.hidden_size / config.num_attention_heads,
            quantization=qconfig,
        )

        trt_config.mapping.gpus_per_node = min(trt_config.mapping.world_size, 8)

        return trt_config

    def get_plugins_config(self) -> PluginConfig:
        config = super().get_plugins_config()
        config.moe_plugin = "disable"  # TODO : Mixtral?
        config.bert_attention_plugin = "disable"
        config.gpt_attention_plugin = self.dtype
        config.gemm_plugin = self.dtype

        return config

    @staticmethod
    def supports_strong_typing() -> bool:
        return False


class MistralForCausalLM(CausalLM, HuggingFaceHubModel):
    MODEL_CONFIG = MistralConfig
    HF_LIBRARY_TARGET_MODEL_CLASS = TransformersMistralForCausalLM
    TRT_LLM_TARGET_MODEL_CLASS = LLaMAForCausalLM

    @staticmethod
    def convert_weights(
        target: PretrainedModel,
        source: TransformersPretrainedModel,
        config: PretrainedConfig,
    ) -> Dict[str, np.ndarray]:
        if config.quant_mode.has_any_quant():
            raise NotImplementedError("Quantization is not supported yet.")

        return load_from_hf_llama(target, source, config.mapping, config.dtype)
